# Generated by Django 3.0.1 on 2019-12-19 17:59

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('list_based_flavorpack', '0006_auto_20191212_1155'),
    ]

    operations = [
        migrations.AlterField(
            model_name='listbasedregion',
            name='machine_name',
            field=models.CharField(help_text='A url/api-friendly unique key for use in API calls and such.', max_length=100, unique=True, validators=[django.core.validators.RegexValidator('^[A-Za-z0-9\\-\\_]*$', 'Only numbers and letters are allowed. Invalid API Key.')], verbose_name='API Key'),
        ),
    ]



'''
    Copyright 2018 SerialLab, CORP

    This file is part of ListBasedFlavorpack.

    RandomFlavorpack is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RandomFlavorpack is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with RandomFlavorpack.  If not, see <http://www.gnu.org/licenses/>.
'''
from rest_framework import status
from serialbox.rules.common import PreprocessingRule
from serialbox.rules.errors import RuleError


class OddNumberError(RuleError):

    def __init__(self, detail=None):
        self.default_detail = ('No odd number requests are allowed! '
                               'To disable this remove the OddNumberError '
                               'rule from the GENERATOR_PREPROCESSING_RULES settings.')
        # NOTE:
        # the status is 400 by default, just including here as an example!!!
        self.status_code = status.HTTP_400_BAD_REQUEST
        RuleError.__init__(self, detail=detail)


class NoOddRequests(PreprocessingRule):
    '''
    Example rule that doesn't allow requests for odd numbers.
    '''

    def execute(self, request, pool, region, size):
        # check the size
        if size % 2 != 0:
            raise OddNumberError()



'''
    Copyright 2019 SerialLab, CORP

    This file is part of ListBasedFlavorPack.

    RandomFlavorpack is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RandomFlavorpack is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with RandomFlavorpack.  If not, see <http://www.gnu.org/licenses/>.
'''
import coreapi, coreschema
from rest_framework.schemas import ManualSchema
from rest_framework.views import APIView
from rest_framework.response import Response
from list_based_flavorpack.models import ListBasedRegion
from list_based_flavorpack.utils import clone_list_based_pool



class CloneListBasedFlavorpackView(APIView):
    """
    Clones a list based flavorpack using the same logic as the
    internal management command.
    """
    queryset = ListBasedRegion.objects.none()

    schema = ManualSchema(fields=[
        coreapi.Field(
            "machine_name",
            required=True,
            location="path",
            schema=coreschema.String()
        ),
        coreapi.Field(
            "new_machine_name",
            required=True,
            location="path",
            schema=coreschema.String()
        ),
    ])

    def post(self, request, machine_name=None, new_machine_name=None):
        if machine_name and new_machine_name:
            clone_list_based_pool(machine_name, new_machine_name)
            return Response("The pool was copied.")



'''
    Copyright 2018 SerialLab, CORP

    This file is part of ListBasedFlavorpack.

    RandomFlavorpack is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RandomFlavorpack is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with RandomFlavorpack.  If not, see <http://www.gnu.org/licenses/>.
'''
from serialbox.api.viewsets import FormMixin
from serialbox import viewsets

from list_based_flavorpack.models import ListBasedRegion, ProcessingParameters
from list_based_flavorpack.api import serializers


class ListBasedRegionViewSet(viewsets.SerialBoxModelViewSet, FormMixin):
    queryset = ListBasedRegion.objects.all()
    lookup_field = 'machine_name'

    def get_serializer_class(self):
        '''
        Return a different serializer depending on the client request.
        '''
        ret = serializers.ListBasedRegionSerializer
        try:
            if self.request.query_params.get('related') == 'true':
                ret = serializers.HyperlinkedListBasedRegionSerializer
        except AttributeError:
            pass
        return ret


class ProcessingParametersViewSet(viewsets.SerialBoxModelViewSet, FormMixin):
    queryset = ProcessingParameters.objects.all()
    serializer_class = serializers.ProcessingParametersSerializer


    
list_based_region_list = ListBasedRegionViewSet.as_view({
    'get': 'list'
})
list_based_region_create = ListBasedRegionViewSet.as_view({
    'post': 'create'
})
list_based_region_detail = ListBasedRegionViewSet.as_view({
    'get': 'retrieve'
})
list_based_region_modify = ListBasedRegionViewSet.as_view({
    'put': 'partial_update',
    'delete': 'destroy'
})
list_based_region_form = ListBasedRegionViewSet.as_view({
    'get': 'form'
})



'''
    Copyright 2018 SerialLab, CORP

    This file is part of ListBasedFlavorpack.

    RandomFlavorpack is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RandomFlavorpack is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with RandomFlavorpack.  If not, see <http://www.gnu.org/licenses/>.
'''

from serialbox.flavor_packs import FlavorPackApp


class ListBasedFlavorpackConfig(FlavorPackApp):
    # The flavorpack app is a django AppConfig instance
    # with a meta-class that does some auto-registration for flavorpacks...
    # for more on AppConfigs see the Django documentation.
    name = 'list_based_flavorpack'
    verbose_name = 'List Based Number Regions'

    @property
    def pool_slug_fields(self):
        return {
            'listbasedregion_set':
            'list_based_flavorpack.api.serializers.listbasedregion_set'
        }

    # Each flavorpack must supply hyperlink fields for it's serializer fields
    @property
    def pool_hyperlink_fields(self):
        return {
            'listbasedregion_set':
            'list_based_flavorpack.api.serializers.listbasedregion_hyperlink_set'
        }
    # Each flavorpack must supply number generators for it's regions

    @property
    def generators(self):
        # here we define the model and generator that
        # is responsible for creating words for the API and maintaining
        # the appropriate state in the database
        return {
            'list_based_flavorpack.models.ListBasedRegion':
            'list_based_flavorpack.generators.list_based.ListBasedGenerator'
        }

    @property
    def api_urls(self):
        # this is the name of the URL config that returns word regions,
        # returning it makes it visible on the self-documenting
        # django rest framework HTML interface at the Root Api page...
        return ['list-based-region-list', 'list-based-region-create']



'''
    Copyright 2018 SerialLab, CORP

    This file is part of ListBasedFlavorPack.

    RandomFlavorpack is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RandomFlavorpack is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with RandomFlavorpack.  If not, see <http://www.gnu.org/licenses/>.
'''

import os
import uuid
from django.db import models
from django.utils.translation import gettext_lazy as _
from threading import Lock
from haikunator import Haikunator

from serialbox import models as sb_models
from quartet_output.models import EndPoint, AuthenticationInfo
from quartet_capture.models import Rule
from quartet_capture.haiku import nouns, adjectives
from quartet_templates.models import Template

haiku = Haikunator(adjectives=adjectives, nouns=nouns)


def haikunate():
    '''
    Since the haikunator is a class method
    it could not be used directly as a default callable for
    a django field...hence this function.
    '''
    lock = None
    try:
        lock = Lock()
        lock.acquire()
        ret = haiku.haikunate(token_length=8, token_hex=True,
                              delimiter='_').replace('-', '')
    finally:
        if lock:
            lock.release()
    return ret


class ListBasedRegion(sb_models.Region):
    '''
    The List-based region defines a region that retrieves numbers from
    a third-party system/application and uses a file containing a list
    for its numbers. It is tied to an Endpoint and AuthenticationInfo
    from the QU4RTET output app.
    '''
    last_number_line = models.IntegerField(
        verbose_name=_('Last Number Line'),
        help_text=_('The line number of the last number issued '
                    'maintains the state of the list-based region.'),
        default=1)
    end_point = models.ForeignKey(
        EndPoint,
        on_delete=models.SET_NULL,
        null=True, blank=True,
        verbose_name=_("End Point"),
        help_text=_("A protocol-specific endpoint defining where"
                    "any data will come from"), )
    rule = models.ForeignKey(
        Rule,
        null=True, blank=True,
        on_delete=models.SET_NULL,
        verbose_name=("Processing Rule"),
        help_text=_(
            "A rule that may be executed by the region processing class."))
    authentication_info = models.ForeignKey(
        AuthenticationInfo,
        null=True, blank=True,
        on_delete=models.SET_NULL,
        verbose_name=_("Authentication Info"),
        help_text=_("The Authentication Info to use."))
    template = models.ForeignKey(
        Template,
        null=True, blank=True,
        on_delete=models.SET_NULL,
        verbose_name=_("Message Template"),
        help_text=_(
            "The Django/Jinja template to send a formatted request for number ranges")
    )
    processing_class_path = models.CharField(
        max_length=150,
        null=False,
        help_text=_(
            'The full python path to the class that will be processing region allocations'),
        verbose_name=_('Processing Class Path'),
        default='list_based_flavorpack.processing_classes.third_party_processing.processing.ThirdPartyProcessingClass'
    )
    file_id = models.UUIDField(
        default=uuid.uuid1,
        help_text=_('If this range utilizes list-based files, then this will'
                    'be the name of the file located in the directory path.'
                    ),
        verbose_name=_('File Name')
    )
    database_name = models.CharField(
        max_length=50, null=True, blank=True,
        default=haikunate,
        help_text=_('The name of the database file if this is a database '
                    'range.'),
        verbose_name=('Database File')
    )
    directory_path = models.CharField(
        max_length=150,
        default="/var/quartet/numbers",
        null=True, blank=True,
        help_text=_(
            "The full path to the directory where numbers will be stored. "
            "Leave blank for default. Default is /var/quartet/numbers"))
    number_replenishment_size = models.IntegerField(
        verbose_name=_("Number Replenishment Size"),
        null=False,
        blank=False,
        help_text=_(
            "The size that the outbound message will request from the third-party system, "
            "if numbers available are low. E.g.: 500"),
        default=5000
    )

    @property
    def file_path(self):
        '''
        The full path to the file.
        '''
        return os.path.join(self.directory_path, str(self.file_id))

    @property
    def db_file_path(self):
        '''
        If the storage mechanism is a database file, return that.
        :return:
        '''
        return os.path.join(self.directory_path, '%s.%s' %
                            (self.database_name, 'db'))


class ProcessingParameters(models.Model):
    '''
    A key-value pair object meant to hold parameters used for processing classes.
    '''
    list_based_region = models.ForeignKey(ListBasedRegion, null=True,
                                          blank=True, on_delete=models.CASCADE,
                                          related_name="processing_parameters",
                                          verbose_name=_(
                                              "Processing Class Parameter"),
                                          help_text=_(
                                              "A key-value pair object meant to hold parameters "
                                              "used for processing classes."))
    key = models.CharField(max_length=200,
                           null=False,
                           help_text=_(
                               "The key part of the key-value pair. Example: q"))
    value = models.CharField(max_length=400,
                             null=True, blank=True,
                             help_text=_(
                                 "The value part of the key-value pair."))

    def __str__(self):
        return "{\"%s\": \"%s\"}" % (self.key, self.value)



# -*- coding: utf-8
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.

from __future__ import unicode_literals, absolute_import
import os
import django

DEBUG = True
USE_TZ = True

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "u_6k^p*oqxbq*h+sbm(kc=iqw&vk6_@gp#!d%p=zqh7a&(^=_o"
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}
# DATABASES = {
#     "default": {
#         "ENGINE": "django.db.backends.sqlite3",
#         "NAME": ":memory:",
#         "TEST_NAME": os.path.join(os.path.dirname(__file__), 'test.db'),
#     }
# }
ROOT_URLCONF = "tests.urls"

REST_FRAMEWORK = {
    'DEFAULT_PERMISSION_CLASSES': (
        'rest_framework.permissions.IsAuthenticated',
        'rest_framework.permissions.DjangoModelPermissions'
    ),
}

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sites",
    "quartet_epcis.apps.QuartetEPCISConfig",    
    "quartet_output.apps.QuartetOutputConfig",
    "quartet_capture.apps.QuartetCaptureConfig",
    "quartet_templates.apps.QuartetTemplatesConfig",
    'rest_framework',
    'rest_framework.authtoken',    
    "serialbox.apps.PoolsConfig",
    "list_based_flavorpack.apps.ListBasedFlavorpackConfig"
]

SITE_ID = 1

if django.VERSION >= (1, 10):
    MIDDLEWARE = ()
else:
    MIDDLEWARE_CLASSES = ()

MEDIA_ROOT = '/tmp/'
#
# import logging
# logging.basicConfig(
#     level=logging.DEBUG,
#     handlers=[logging.StreamHandler()]
# )

# TEMPLATE CONFIGURATION
# ------------------------------------------------------------------------------
# See: https://docs.djangoproject.com/en/dev/ref/settings/#templates
TEMPLATES = [
    {
        # See: https://docs.djangoproject.com/en/dev/ref/settings/#std:setting-TEMPLATES-BACKEND
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        # See: https://docs.djangoproject.com/en/dev/ref/settings/#template-dirs
        'DIRS': [],
        'OPTIONS': {
            # See: https://docs.djangoproject.com/en/dev/ref/settings/#template-debug
            'debug': DEBUG,
            # See: https://docs.djangoproject.com/en/dev/ref/settings/#template-loaders
            # https://docs.djangoproject.com/en/dev/ref/templates/api/#loader-types
            'loaders': [
                'django.template.loaders.filesystem.Loader',
                'django.template.loaders.app_directories.Loader',
            ],
            # See: https://docs.djangoproject.com/en/dev/ref/settings/#template-context-processors
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.template.context_processors.i18n',
                'django.template.context_processors.media',
                'django.template.context_processors.static',
                'django.template.context_processors.tz',
                'django.contrib.messages.context_processors.messages',
                # Your stuff: custom template context processors go here
            ],
        },
    },
]



# Generated by Django 2.1.1 on 2018-10-02 15:52

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('list_based_flavorpack', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='processingparameters',
            name='list_based_region',
            field=models.ForeignKey(help_text='A key-value pair object meant to hold parameters used for processing classes.', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='processing_parameters', to='list_based_flavorpack.ListBasedRegion', verbose_name='Processing Class Parameter'),
        ),
    ]



'''
    Copyright 2018 SerialLab, CORP

    This file is part of ListBasedFlavorpack.

    RandomFlavorpack is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RandomFlavorpack is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with RandomFlavorpack.  If not, see <http://www.gnu.org/licenses/>.
'''

from rest_framework import serializers as rf_serializers
from serialbox.api.serializers import RegionSerializer
from list_based_flavorpack import models



##
# These fields will be added to the pool serializer automatically since
# they are specified in the pool_slug_fields and the pool_hyperlinked_fields
# values in the words_flavorpack.apps.WordsFlavorpackConfig module.
##
listbasedregion_set = rf_serializers.SlugRelatedField(
    many=True,
    queryset=models.ListBasedRegion.objects.all(),
    slug_field='machine_name',
    required=False
)

listbasedregion_hyperlink_set = rf_serializers.HyperlinkedRelatedField(
    many=True,
    read_only=True,
    view_name='list-based-region',
    lookup_field='machine_name',
)



class ProcessingParametersSerializer(rf_serializers.ModelSerializer):
    '''
    Default serializer for Processing Parameters
    '''
    id = rf_serializers.IntegerField(write_only=False)
    
    class Meta:
        model = models.ProcessingParameters
        fields = ('id', 'key', 'value', 'list_based_region',)


class ListBasedRegionSerializer(RegionSerializer):
    '''
    Specifies the model...
    '''
    processing_parameters = ProcessingParametersSerializer(many=True, read_only=False)
    
    class Meta(object):
        model = models.ListBasedRegion
        fields = ('id', 'pool', 'created_date', 'modified_date', 'readable_name',
                  'machine_name', 'active', 'order', 'last_number_line',
                  'number_replenishment_size', 'directory_path',
                  'processing_class_path', 'end_point', 'rule', 'authentication_info',
                  'template', 'processing_parameters')

    def create(self, validated_data):
        '''
        Creates a list-based region and its processing params.
        '''
        processing_parameters = validated_data.pop("processing_parameters")
        region = models.ListBasedRegion.objects.create(**validated_data)
        for processing_parameter in processing_parameters:
            models.ProcessingParameters.objects.create(list_based_region=region, **processing_parameter)
        return region
    
    def update(self, instance, validated_data):
        '''
        Creates a list-based region and its processing params.
        '''
        processing_parameters = validated_data.pop("processing_parameters")
        super(ListBasedRegionSerializer, self).update(instance, validated_data)
        posted_ids = set(param.get('id', None) for param in processing_parameters)
        existing_params = set(models.ProcessingParameters.objects.filter(list_based_region=instance).values_list('id', flat=True))
        ids_for_removal = existing_params - posted_ids
        objects_for_removal = models.ProcessingParameters.objects.filter(list_based_region=instance, pk__in=ids_for_removal)
        objects_for_removal.delete()
        for processing_parameter in processing_parameters:
            id = processing_parameter.get('id', None)
            if id:
                item = models.ProcessingParameters.objects.get(id=id)
                item.key = processing_parameter.get('key', item.key)
                item.value = processing_parameter.get('value', item.value)
                item.save()
            else:
                models.ProcessingParameters.objects.create(list_based_region=instance, **processing_parameter)
        return instance

class HyperlinkedListBasedRegionSerializer(ListBasedRegionSerializer):
    '''
    Addst the hyperlinked field...
    '''
    pool = rf_serializers.HyperlinkedRelatedField(view_name='pool-detail',
                                                  read_only=True,
                                                  lookup_field='machine_name')



import os
import linecache
import sqlite3
from list_based_flavorpack.processing_classes.third_party_processing import \
    rules
from list_based_flavorpack.models import ListBasedRegion
from list_based_flavorpack.processing_classes.third_party_processing.rules import \
    get_region_table


class ThirdPartyProcessingClass:
    '''
    An example of a processing  class for a list based pool/region.
    Returns UUIDs stored in memory.
    '''

    def __init__(self):
        '''
        Sets processing rules.
        '''
        self.pre_processing_rules = [rules.ValidNumberDirectory,
                                     rules.SufficientNumbersStorage]
        self.post_processing_rules = []

    def get_pre_processing_rules(self):
        '''
        Returns processing rules to be executed prior to the generator running.
        '''
        return self.pre_processing_rules

    def get_post_processing_rules(self):
        '''
        Returns processing rules to be executed after the generator has run.
        '''
        return self.post_processing_rules

    def get_items(self, region, size):
        '''
        Pull numbers from the file created/updated from rules previously executed.
        '''
        first = region.last_number_line
        max_line = first + size
        lines = []
        linecache.checkcache(filename=region.file_path)
        for lineno in range(first, max_line):
            item = linecache.getline(region.file_path, lineno).strip()
            if not item:
                raise ValueError(
                    "There are not enough numbers available for this iteration starting on line number: %s",
                    lineno)
            lines.append(item)
        region.last_number_line = max_line
        region.save()
        return lines


class DBProcessingClass(ThirdPartyProcessingClass):
    '''
    Uses a SQLite database instead of a flat file.  When numbers are used
    they are removed from the database.
    '''

    def __init__(self):
        '''
        Sets processing rules.
        '''
        self.pre_processing_rules = [rules.ValidNumberDirectory,
                                     rules.SufficientDBNumbers]
        self.post_processing_rules = []

    def get_items(self, region: ListBasedRegion, size):
        connection = sqlite3.connect(region.db_file_path)
        cursor = connection.cursor()
        cursor.execute(
            "SELECT serial_number FROM %s WHERE used = 0 LIMIT ?" % get_region_table(region),
            (size,)
        )
        rows = cursor.fetchall()
        lines = []
        for row in rows:
            lines.append(row[0])
            cursor.execute('DELETE FROM %s WHERE serial_number = ?'
                           % get_region_table(region), (row[0],))
        connection.commit()
        if len(rows) < size:
            raise ValueError("There are not enough numbers to satisfy the "
                             "request.")
        return lines
